# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.api.job_input import IProperties


class CachedPropertiesWrapper(IProperties):
    """
    Wraps any IProperties so that get_* calls are cached until set_* is called
    Needed mainly because .sql file execution performs SQL argument substitution. In the general case this means
    invoking HDFS read command, however properties are rarely updated, so a cache reduces HDFS access, thus reducing
    overall job execution time.
    """

    def __init__(self, properties_impl: IProperties):
        self.provider = properties_impl
        self.cached_dict = (
            None  # when None then cache needs to be refreshed on next get_*()
        )

    def get_property(self, name, default_value=None):
        res = self._get_cached_dict().get(name, default_value)
        return res

    def get_all_properties(self):
        return self._get_cached_dict()

    def set_all_properties(self, properties):
        self.provider.set_all_properties(properties)
        self.cached_dict = None

    def _get_cached_dict(self):
        if self.cached_dict is None:
            self.cached_dict = self.provider.get_all_properties()
        return self.cached_dict



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import os
import pathlib
from typing import cast
from typing import Dict
from typing import Optional

import click
from vdk.internal.builtin_plugins.run import job_input_error_classifier
from vdk.internal.builtin_plugins.run.data_job import DataJobFactory
from vdk.internal.builtin_plugins.run.execution_tracking import (
    ExecutionTrackingPlugin,
)
from vdk.internal.builtin_plugins.version import version
from vdk.internal.core import errors
from vdk.internal.core.context import CoreContext

log = logging.getLogger(__name__)


class CliRunImpl:
    def __init__(self, job_factory: DataJobFactory = None):
        self.__job_factory = job_factory if job_factory else DataJobFactory()

    @staticmethod
    def __validate_and_parse_args(arguments: str) -> Optional[Dict]:
        try:
            if arguments:
                return json.loads(arguments)
            else:
                return None
        except Exception as e:
            blamee = errors.ResolvableBy.USER_ERROR
            errors.log_and_rethrow(
                blamee,
                logging.getLogger(__name__),
                what_happened="Failed to validate job arguments.",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                consequences=errors.MSG_CONSEQUENCE_TERMINATING_APP,
                countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION,
                exception=e,
                wrap_in_vdk_error=True,
            )

    def create_and_run_data_job(
        self,
        context: CoreContext,
        data_job_directory: pathlib.Path,
        arguments: Optional[str],
    ):
        log.info(f"Run job with directory {data_job_directory}")
        context.plugin_registry.load_plugin_with_hooks_impl(ExecutionTrackingPlugin())

        job = self.__job_factory.new_datajob(
            data_job_directory=data_job_directory, core_context=context
        )
        args = self.__validate_and_parse_args(arguments)

        execution_result = None
        try:
            execution_result = job.run(args)
            log.info(f"Data Job execution summary: {execution_result}")
        except BaseException as e:
            errors.log_and_rethrow(
                job_input_error_classifier.whom_to_blame(
                    e, __file__, data_job_directory
                ),
                log,
                what_happened="Failed executing job.",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                consequences=errors.MSG_CONSEQUENCE_TERMINATING_APP,
                countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION
                + " Most likely a prerequisite or plugin of one of the key VDK components failed, see"
                + " logs for details and ensure the prerequisite for the failed component (details in stacktrace).",
                exception=e,
            )
        if execution_result.is_failed() and execution_result.get_exception_to_raise():
            raise execution_result.get_exception_to_raise()


@click.command(
    help="Run a Data Job. "
    """
     Examples:

     \b
     # This will run the Data Job from directory example-job (it takes a few minutes)
     vdk run /home/user/data-jobs/example-job

"""
)
@click.argument(
    "data_job_directory",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
)
@click.option(
    "--arguments",
    type=click.STRING,
    required=False,
    help="Pass arguments. "
    "Those arguments will be passed to each step. "
    "Must be in valid JSON format",
)
@click.pass_context
def run(ctx: click.Context, data_job_directory: str, arguments: str) -> None:
    """
    Entry point of the CLI run. It start a run (execution) of a data job.
    """
    log.info(
        f"Versatile Data Kit (VDK){os.linesep}{version.get_version_info()}{os.linesep + '-' * 80}"
    )
    context: CoreContext = cast(CoreContext, ctx.obj)
    run_impl = CliRunImpl()
    run_impl.create_and_run_data_job(
        context, pathlib.Path(data_job_directory), arguments
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import types
from sys import modules
from typing import cast

from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.config import vdk_config
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core import errors
from vdk.internal.core.statestore import CommonStoreKeys


def configure_loggers(
    job_name: str = "",
    attempt_id: str = "no-id",
    log_config_type: str = None,
    vdk_logging_level: str = "DEBUG",
) -> None:
    """
    Configure default logging configuration

    :param job_name: the name of the job
    :param attempt_id: the id of the current job run attempt
    :param log_config_type: where the job is executed: CLOUD or LOCAL
    :param vdk_logging_level: The level for vdk specific logs.
    """

    import logging.config

    if _is_already_configured():
        log = logging.getLogger(__name__)
        msg = "Logging already configured. This seems like a bug. Fix me!."
        # PLugins would override loggin directly wiht logging library not through this method
        # it is probably OK for tests to configure logging twice. VDK, however, shouldn't.
        log.warning(msg)

    # Likely most logging system (like Log Insight) show the hostname from where the syslog message arrived so no need to include it here.
    DETAILED_FORMAT = (
        f"%(asctime)s [VDK] {job_name} [%(levelname)-5.5s] %(name)-30.30s %(filename)20.20s:%("
        f"lineno)-4.4s %(funcName)-16.16s[id:{attempt_id}]- %(message)s"
    )

    _LOGGERS = {
        "requests_kerberos": {"level": "INFO"},
        "requests_oauthlib": {"level": "INFO"},
        "urllib3": {"level": "INFO"},
        "vdk": {"level": vdk_logging_level},
    }

    _FORMATTERS = {"detailedFormatter": {"format": DETAILED_FORMAT}}

    _CONSOLE_HANDLER = {
        "class": "logging.StreamHandler",
        "formatter": "detailedFormatter",
        "stream": "ext://sys.stderr",
    }

    if "CLOUD" == log_config_type:
        CLOUD = {  # @UnusedVariable
            "version": 1,
            "handlers": {"consoleHandler": _CONSOLE_HANDLER},
            "formatters": _FORMATTERS,
            "root": {"handlers": ["consoleHandler"]},
            "loggers": _LOGGERS,
            "disable_existing_loggers": False,
        }
        logging.config.dictConfig(CLOUD)
    elif "NONE" == log_config_type:
        pass
    else:
        LOCAL = {  # @UnusedVariable
            "version": 1,
            "handlers": {"consoleHandler": _CONSOLE_HANDLER},
            "formatters": _FORMATTERS,
            "root": {"handlers": ("consoleHandler",)},
            "loggers": _LOGGERS,
            "disable_existing_loggers": False,
        }
        logging.config.dictConfig(LOCAL)
    _set_already_configured()


def _set_already_configured() -> None:
    setattr(modules[__name__], "logging_already_configured", True)


def _is_already_configured() -> bool:
    res = hasattr(modules[__name__], "logging_already_configured")
    return res


def ensure_logging_for_current_test() -> None:
    """
    To be called from test code only.

    Ensures that logging is configured, either by another class, or by the caller.
    Useful when debugging a single test class, not the whole suite (assuming all suites configure logging)

    Uses filename of caller for derived_dataset parameter, e.g. if this method is called from impala_connection_test.py
    then logs would be configured to look like:

    2018-04-19 16:05:48,488=1524143148[VDK] impala_connection_test.py [DEBUG]  ......
    """
    if not _is_already_configured():
        try:
            from inspect import currentframe

            me = cast(types.FrameType, currentframe())
            caller = cast(types.FrameType, me.f_back)
            path = caller.f_locals["__file__"]
            from os.path import split

            dir, fname = split(path)  # @ReservedAssignment
        except:
            fname = "n/a"
        from time import time

        configure_loggers(fname, str(time()))


class LoggingPlugin:
    """
    Define the logging plugin
    """

    @hookimpl
    def initialize_job(self, context: JobContext) -> None:
        """
        Initialize logging for running Data Job.

        When running a data job logging is more verbose, it includes more details like op id, job name.
        """

        attempt_id = context.core_context.state.get(CommonStoreKeys.ATTEMPT_ID)
        job_name = context.name
        log_config_type = context.core_context.configuration.get_value(
            vdk_config.LOG_CONFIG
        )
        vdk_log_level = context.core_context.configuration.get_value(
            vdk_config.LOG_LEVEL_VDK
        )
        try:  # If logging initialization fails we want to attempt sending telemetry before exiting VDK
            configure_loggers(
                job_name,
                attempt_id,
                log_config_type=log_config_type,
                vdk_logging_level=vdk_log_level,
            )
            log = logging.getLogger(__name__)
            log.debug(f"Initialized logging for log type {log_config_type}.")
        except Exception as e:
            # Have local logs on DEBUG level, when standard log configuration fails
            logging.basicConfig(level=logging.DEBUG)
            log = logging.getLogger(__name__)
            log.warning(
                "Logging configuration %s failed.\n"
                "Trying to send telemetry for Job attempt: %s"
                % (log_config_type, attempt_id)
            )
            errors.log_and_rethrow(
                errors.find_whom_to_blame_from_exception(e),
                log=logging.getLogger(__name__),
                what_happened="Failed to initialize logging",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                consequences="Logging is critical VDK component. VDK will now exit.",
                countermeasures="Depending on stacktrace",
                exception=e,
            )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pluggy

"""
The plugin project name. It is used in auto-discovery of the hooks and plugins.
If external module needs to recognize as a plugin module they need to specify this as an entry point.
"""
GROUP_NAME = "vdk.plugin.run"

"""
Decorator helper class for marking functions as hook specifications.

The hook spec can be configured using following variables:

If ``firstresult`` is ``True`` the 1:N hook call (N being the number of registered
hook implementation functions) will stop at I<=N when the I'th function
returns a non-``None`` result.

If ``historic`` is ``True`` calls to a hook will be memorized and replayed
on later registered plugins.
"""
hookspec = pluggy.HookspecMarker(GROUP_NAME)

"""
Decorator helper class for marking functions as hook implementations.
The method name must match one of those defined as hookspec
The method signature - the arguments need to be subset of arguments defined in hookspec

Plugin execution order can be configured in the decorator with following variables:

If ``optionalhook`` is ``True`` a missing matching hook specification will not result
in an error (by default it is an error if no matching spec is found).

If ``tryfirst`` is ``True`` this hook implementation will run as early as possible
in the chain of N hook implementations for a specification.

If ``trylast`` is ``True`` this hook implementation will run as late as possible
in the chain of N hook implementations.

If ``hookwrapper`` is ``True`` the hook implementations needs to execute exactly
one ``yield``.  The code before the ``yield`` is run early before any non-hookwrapper
function is run.  The code after the ``yield`` is run after all non-hookwrapper
function have run.  The ``yield`` receives a :py:class:`.callers._Result` object
representing the exception or result outcome of the inner calls (including other
hookwrapper calls).
Example:

@hookimpl(hookwrapper=True)
def func():
    before_other_hooks_execute()
    outcome = yield
    # outcome.excinfo may be None or a (cls, val, tb)
    # outcome.get_result() will raise exception if there was an exception before.
    after_other_hooks_executed(outcome.get_result())
    outcome.force_result(new_res)  # set new return result

"""
hookimpl = pluggy.HookimplMarker(GROUP_NAME)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import textwrap
from enum import Enum
from typing import List

import click
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.version import version
from vdk.internal.builtin_plugins.version.new_version_check import Package
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext

log = logging.getLogger(__name__)


class ConfigKey(str, Enum):
    PACKAGE_NAME = "PACKAGE_NAME"
    PACKAGE_INDEX = "PACKAGE_INDEX"
    VERSION_CHECK_PLUGINS = "VERSION_CHECK_PLUGINS"


def new_package(package_name: str, package_index: str) -> Package:
    return Package(package_name, package_index)


class NewVersionCheckPlugin:
    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        config_builder.add(
            key=ConfigKey.PACKAGE_NAME,
            default_value="vdk-core",
            description="Set distribution package name of the library to check for new version",
        )
        config_builder.add(
            key=ConfigKey.PACKAGE_INDEX,
            default_value="https://pypi.org",
            description="Set distribution package name of the library to check for new version",
        )
        config_builder.add(
            key=ConfigKey.VERSION_CHECK_PLUGINS,
            default_value=True,
            description="Set to true if plugins should be checked for new version otherwise false",
        )

    @hookimpl
    def vdk_exit(self, context: CoreContext) -> None:
        try:
            package_list = []
            cfg = context.configuration

            package_name = cfg.get_value(ConfigKey.PACKAGE_NAME)
            package_index = cfg.get_value(ConfigKey.PACKAGE_INDEX)
            if new_package(package_name, package_index).check():
                package_list.append(package_name)

            if cfg.get_value(ConfigKey.VERSION_CHECK_PLUGINS):
                log.debug("Will check for newer versions for all installed plugins.")
                for dist_name, _ in version.list_installed_plugins():
                    if new_package(dist_name, package_index).check():
                        package_list.append(dist_name)

            self._check_version(package_list, package_index)
        except Exception as e:
            log.debug(
                f"Could not check for new version release. "
                f"Error was {e}. We are ignoring the error."
            )

    @staticmethod
    def _check_version(package_list: List[str], package_index: str) -> None:
        """
        Prints out a new version message for the listed packages.

        :param package_list: list of package names
        :param package_index: package index included in the printed `pip install` command
        """
        not_single_package = len(package_list) > 1

        # if package index is not specified that we fetch it from pip repo (pypi.org)
        extra_index_if_needed = (
            f"--extra-index-url {package_index}"
            if (package_index and (package_index != "https://pypi.org"))
            else ""
        )
        pip_command = "pip install --upgrade-strategy eager -U"
        main_message = f"New version{'s' if not_single_package else ''} for {', '.join(package_list)} {'are' if not_single_package else 'is'} available."
        install_message = (
            f"{pip_command} {' '.join(package_list)} {extra_index_if_needed}"
        )
        # We are using eager strategy so that if there's new version of a plugin that satisfies the requirements
        # of the main package we want to upgrade to it.
        click.echo(
            f"""
    ******************************************************************************************

    {textwrap.fill(main_message, 80, subsequent_indent="    ")}

    Please update to latest version by using:
    {install_message}

    ******************************************************************************************
            """,
            err=True,
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import types
from typing import Any
from typing import cast
from typing import Collection
from typing import Container
from typing import Optional

from vdk.api.plugin.connection_hook_spec import (
    ConnectionHookSpec,
)
from vdk.internal.builtin_plugins.connection.decoration_cursor import DecorationCursor
from vdk.internal.builtin_plugins.connection.decoration_cursor import ManagedOperation
from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Cursor
from vdk.internal.builtin_plugins.connection.recovery_cursor import RecoveryCursor
from vdk.internal.core import errors


class ManagedCursor(PEP249Cursor):
    """
    PEP249 Cursor
    """

    def __init__(
        self,
        cursor: Any,
        log: logging.Logger = None,
        connection_hook_spec: ConnectionHookSpec = None,
    ):
        if not log:
            log = logging.getLogger(__name__)
        super().__init__(cursor, log)
        self.__connection_hook_spec = connection_hook_spec

    def __getattr__(self, attr):
        """
        Dynamic interception and delegation of any (non-overridden) attribute access.
        In case an attribute is not explicitly managed (customized by overriding e.g. execute()) -
        this attribute is looked up then the call is delegated, ensuring default behaviour success path.

        First, the non-managed attribute call is redirected to the wrapped native cursor if attribute available,
        otherwise to the superclass if attribute is present.
        If the attribute is not specified by both the native cursor nor the superclass, an AttributeError is raised.

        Default behaviour availability unblocks various ManagedCursor usages that rely on
        currently not explicitly defined in the scope of ManagedCursor attributes.
        E.g. SQLAlchemy dependency that uses the ManagedCursor, does require some specific attributes available.

        For more details on customizing attributes access, see PEP562.
        """
        # native cursor
        if hasattr(self._cursor, attr):
            if isinstance(getattr(self._cursor, attr), types.MethodType):

                def method(*args, **kwargs):
                    return getattr(self._cursor, attr)(*args, **kwargs)

                return method
            return getattr(self._cursor, attr)
        # superclass
        if hasattr(super(), attr):
            if isinstance(getattr(super(), attr), types.MethodType):

                def method(*args, **kwargs):
                    return getattr(super(), attr)(*args, **kwargs)

                return method
            return getattr(super(), attr)
        raise AttributeError

    def execute(
        self, operation: str, parameters: Optional[Container] = None
    ) -> None:  # @UnusedVariable
        managed_operation, decoration_cursor = (
            ManagedOperation(operation, parameters),
            None,
        )
        if self.__connection_hook_spec:
            if (
                self.__connection_hook_spec.db_connection_validate_operation.get_hookimpls()
            ):
                self._log.debug("Validating query:\n%s" % operation)
                try:
                    self.__connection_hook_spec.db_connection_validate_operation(
                        operation=operation, parameters=parameters
                    )
                except Exception as e:
                    errors.log_and_rethrow(
                        errors.ResolvableBy.USER_ERROR,
                        self._log,
                        what_happened="Validating query FAILED.",
                        why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                        consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                        countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION,
                        exception=e,
                    )

            if (
                self.__connection_hook_spec.db_connection_decorate_operation.get_hookimpls()
            ):
                self._log.debug("Decorating query:\n%s" % operation)
                decoration_cursor = DecorationCursor(
                    self._cursor, self._log, managed_operation
                )

                try:
                    self.__connection_hook_spec.db_connection_decorate_operation(
                        decoration_cursor=decoration_cursor
                    )
                except Exception as e:
                    errors.log_and_rethrow(
                        errors.ResolvableBy.PLATFORM_ERROR,
                        self._log,
                        what_happened="Decorating query FAILED.",
                        why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                        consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                        countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION,
                        exception=e,
                    )

        self._log.info("Executing query:\n%s" % managed_operation.get_operation())
        try:
            super().execute(*managed_operation.get_operation_parameters_tuple())
            self._log.info("Executing query SUCCEEDED.")
        except Exception as e:
            try:
                self._recover_operation(e, managed_operation)
            except Exception as e:
                # todo: error classification
                # if job_input_error_classifier.is_user_error(e):
                blamee = errors.ResolvableBy.USER_ERROR
                # else:
                #     blamee = errors.ResolvableBy.PLATFORM_ERROR
                errors.log_and_rethrow(
                    blamee,
                    self._log,
                    what_happened="Executing query FAILED.",
                    why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                    consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                    countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION,
                    exception=e,
                )

    def fetchall(self) -> Collection[Collection[Any]]:
        self._log.info("Fetching all results from query ...")
        try:
            res = self._cursor.fetchall()
            self._log.info("Fetching all results from query SUCCEEDED.")
            return cast(Collection[Collection[Any]], res)
        except:
            self._log.info("Fetching all results from query FAILED.")
            raise

    def close(self) -> None:
        self._log.info("Closing DB cursor ...")
        self._cursor.close()
        self._log.info("Closing DB cursor SUCCEEDED.")

    def _recover_operation(self, exception, managed_operation):
        # TODO: configurable generic re-try.
        if (
            not self.__connection_hook_spec
            or not self.__connection_hook_spec.db_connection_recover_operation.get_hookimpls()
        ):
            raise exception

        recovery_cursor = RecoveryCursor(
            self._cursor,
            self._log,
            exception,
            managed_operation,
            self.__connection_hook_spec.db_connection_decorate_operation,
        )
        self._log.debug(f"Recovery of query {managed_operation.get_operation()}")
        try:
            self.__connection_hook_spec.db_connection_recover_operation(
                recovery_cursor=recovery_cursor
            )
            self._log.debug(
                f"Recovery of query SUCCEEDED "
                f"after {(recovery_cursor.get_retries())} retries."
            )
        except Exception as e:
            self._log.debug(
                f"Recovery of query FAILED "
                f"after {(recovery_cursor.get_retries())} retries."
            )
            if type(e) is type(exception) and e.args == exception.args:  # re-raised
                raise exception
            raise e from exception  # keep track of originating one



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Any
from typing import Callable

from vdk.api.plugin.connection_hook_spec import (
    ConnectionHookSpec,
)
from vdk.internal.builtin_plugins.connection.managed_connection_base import (
    ManagedConnectionBase,
)
from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Connection


class WrappedConnection(ManagedConnectionBase):
    """
    Wrap any DB connection ones (e.g. connection to SAP Hana) into a ManagedConnection
    """

    def __init__(
        self,
        log: logging.Logger,
        new_connection_builder_function: Callable[[], PEP249Connection],
        connection_hook_spec: ConnectionHookSpec,
    ) -> None:
        """
        :param new_connection_builder_function: method that returns a new (e.g. SAP Hana) connection
            Example
                def connection() -> ManagedConnectionBase:
                    db = pyhdb.connect(host='hana-prod-d1.northpole.com', port=30015, user='claus', password='hohoho')
                    return db
        :param connection_hook_spec: connection hook implementations from plugins
        """
        super().__init__(log, None, connection_hook_spec)
        self._log = logging.getLogger(__name__)
        self._new_connection_builder_function = new_connection_builder_function

    def _connect(self) -> Any:
        self._log.debug("Establishing Wrapped connection ...")
        conn = self._new_connection_builder_function()
        self._log.debug(f"Establishing Wrapped connection DONE: {str(conn)}")
        return conn




RELEASE_VERSION='0.1.477708478'
BUILD_DATE='Wed Feb 23 14:54:40 UTC 2022'
BUILD_MACHINE_INFO='Linux runner-opyxurxq-project-28359933-concurrent-1t8tnt 5.4.129-62.227.amzn2.x86_64 #1 SMP Wed Jul 7 00:08:43 UTC 2021 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='2127938987'
GIT_COMMIT_SHA='2816bf43a32bc5976ca8c9510c7593cce3e1944a'
GIT_BRANCH='main'



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Dict
from typing import Union

from vdk.api.job_input import IProperties
from vdk.api.plugin.plugin_input import IPropertiesServiceClient

from .base_properties_impl import check_valid_property

log = logging.getLogger(__name__)


PropertyValue = Union[int, float, str, list, dict, None]


class DataJobsServiceProperties(IProperties):
    """
    Data Jobs Properties implementation.
    """

    __VALID_TYPES = [int, float, str, list, dict, type(None)]

    def __init__(
        self,
        job_name: str,
        team_name: str,
        properties_service_client: IPropertiesServiceClient,
    ):
        log.debug(
            f"Data Job Properties for job {job_name} with service client: {properties_service_client}"
        )
        self._job_name = job_name
        self._team_name = team_name
        self._properties_service_client = properties_service_client

    def get_property(
        self, name: str, default_value: PropertyValue = None
    ) -> PropertyValue:
        """
        :param name: The name of the property
        :param default_value: default value ot return if missing
        """
        properties = self.get_all_properties()
        if name in properties:
            return properties[name]
        else:
            log.warning(
                "Property {} is not among Job properties, returning default value: {}".format(
                    name, default_value
                )
            )
            return default_value

    def get_all_properties(self) -> Dict[str, PropertyValue]:
        """
        :return: all stored properties
        """
        return self._properties_service_client.read_properties(
            self._job_name, self._team_name
        )

    def set_all_properties(self, properties: Dict[str, PropertyValue]) -> None:
        """
        Persists the passed properties overwriting all previous properties
        """
        for (k, v) in list(properties.items()):
            check_valid_property(
                k, v, DataJobsServiceProperties.__VALID_TYPES
            )  # throws

        self._properties_service_client.write_properties(
            self._job_name, self._team_name, properties
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import queue
import sys
import threading
from collections import defaultdict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from vdk.api.job_input import IIngester
from vdk.api.plugin.plugin_input import IIngesterPlugin
from vdk.internal.builtin_plugins.ingestion import ingester_utils
from vdk.internal.builtin_plugins.ingestion.ingester_configuration import (
    IngesterConfiguration,
)
from vdk.internal.builtin_plugins.ingestion.ingester_utils import AtomicCounter
from vdk.internal.core import errors
from vdk.internal.core.errors import PlatformServiceError
from vdk.internal.core.errors import ResolvableBy
from vdk.internal.core.errors import UserCodeError
from vdk.internal.core.errors import VdkConfigurationError

log = logging.getLogger(__name__)


class IngesterBase(IIngester):
    """
    Plugins can subclass this class to provide different ingestion capabilities.
    This can be done by overriding `ingest_payload`, and provide customizations as
    to what data formats can be send, and what methods for sending are to be used.

    Sending internal telemetry for payloads can be facilitated by using `send` to
    send the telemetry payload.

    The thread pools are provided by vdk and can be customized by the plugins.
    """

    def __init__(
        self,
        data_job_name: str,
        op_id: str,
        ingester: IIngesterPlugin,
        ingest_config: IngesterConfiguration,
        pre_processors: Optional[List[IIngesterPlugin]] = None,
        post_processors: Optional[List[IIngesterPlugin]] = None,
    ):
        """
        This constructor must be called by inheritors.

        :param data_job_name: string
            Name of the data job.
        :param op_id: string
            OpId of the data job run.
        :param ingest_config: IngesterConfiguration
            Configuration related to the core ingestion API.
        :param pre_processors: Optional[List[IIngesterPlugin]]
            A list of initialized IIngesterPlugin instances, whose purpose
            is to process the payload before it is ingested.
        :param post_processors: Optional[List[IIngesterPlugin]]
            A list of initialized IIngesterPlugin instances, whose purpose
            is to process the ingestion metadata after the ingestion of the
            payload.
        """
        self._data_job_name = data_job_name
        self._op_id = op_id
        self._ingester = ingester
        self._pre_processors = pre_processors
        self._post_processors = post_processors
        self._number_of_worker_threads = ingest_config.get_number_of_worker_threads()
        self._payload_size_bytes_threshold = (
            ingest_config.get_payload_size_bytes_threshold()
        )
        self._log_upload_errors = ingest_config.get_should_log_upload_errors()

        self._payload_aggregator_timeout_seconds = (
            ingest_config.get_payload_aggregator_timeout_seconds()
        )

        self._fail_count = AtomicCounter()
        self._plugin_errors = defaultdict(
            AtomicCounter
        )  # something like {UserCodeError: 10, VdkConfigurationError: 2}
        self._success_count = AtomicCounter()
        self._closed = AtomicCounter()
        self._objects_queue: queue.Queue = queue.Queue(
            ingest_config.get_objects_queue_size()
        )
        self._payloads_queue: queue.Queue = queue.Queue(
            ingest_config.get_payloads_queue_size()
        )
        self._exception_on_failure = (
            ingest_config.get_should_raise_exception_on_failure()
        )
        self._wait_to_finish_after_every_send = (
            ingest_config.get_wait_to_finish_after_every_send()
        )

        self._start_workers()

    def send_object_for_ingestion(
        self,
        payload: dict,
        destination_table: Optional[str],
        method: Optional[str],
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
    ):
        """
        See parent doc
        """
        self.__verify_payload_format(payload_dict=payload)

        if collection_id is None:
            collection_id = "{data_job_name}|{execution_id}".format(
                data_job_name=self._data_job_name, execution_id=self._op_id
            )

        self._send(
            payload_dict=payload,
            destination_table=destination_table,
            method=method,
            target=target,
            collection_id=collection_id,
        )

        self.__wait_if_necessary()

    def send_tabular_data_for_ingestion(
        self,
        rows: iter,
        column_names: list,
        destination_table: Optional[str],
        method: Optional[str],
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
    ):
        """
        See parent doc
        """
        if len(column_names) == 0 and destination_table is None:
            errors.log_and_throw(
                ResolvableBy.USER_ERROR,
                log,
                what_happened="Failed to ingest tabular data",
                why_it_happened="Either column names or destination table must be specified."
                "Without at least one of those we cannot determine how data should be ingested and we abort.",
                consequences="The data will not be ingested and the current call will fail with an exception.",
                countermeasures="Pass column names or destination table as argument.",
            )

        if not ingester_utils.is_iterable(rows):
            errors.log_and_throw(
                ResolvableBy.USER_ERROR,
                log,
                what_happened="Cannot ingest tabular data",
                why_it_happened=f"The rows argument must be an iterable but it was type: {type(rows)}",
                consequences="The data will not be ingested and current call will fail with an exception.",
                countermeasures="Make sure rows is proper iterator object ",
            )

        if not isinstance(column_names, Iterable):
            errors.log_and_throw(
                ResolvableBy.USER_ERROR,
                log,
                what_happened="Cannot ingest tabular data",
                why_it_happened=f"The column_names argument must be a List (or iterable) but it was: {type(rows)}",
                consequences="The data will not be ingested and current call will fail with an exception.",
                countermeasures="Make sure column_names is proper List object ",
            )

        log.info(
            "Posting for ingestion data for table {table} with columns {columns} against endpoint {endpoint}".format(
                table=destination_table, columns=column_names, endpoint=target
            )
        )

        if collection_id is None:
            collection_id = "{data_job_name}|{execution_id}".format(
                data_job_name=self._data_job_name, execution_id=self._op_id
            )

        # fetch data in chunks to prevent running out of memory
        for page_number, page in enumerate(ingester_utils.get_page_generator(rows)):
            ingester_utils.validate_column_count(page, column_names)
            converted_rows = ingester_utils.convert_table(page, column_names)
            log.debug(
                "Posting page {number} with {size} rows for ingestion.".format(
                    number=page_number, size=len(converted_rows)
                )
            )
            for row in converted_rows:
                self.__verify_payload_format(payload_dict=row)
                self._send(
                    payload_dict=row,
                    destination_table=destination_table,
                    method=method,
                    target=target,
                    collection_id=collection_id,
                )

        self.__wait_if_necessary()

    def _send(
        self,
        payload_dict: dict,
        destination_table: str,
        method: str,
        target: str,
        collection_id: str = None,
    ):
        """
        Send payload to the _objects_queue for processing.)

        :param payload_dict: dict
            Payload to be send to the _objects_queue.
        :param destination_table: string
            The name of the table, where the data should be ingested into.
        :param method: string
            Indicates the ingestion method to be used. E.g.:
                    method="file" -> ingest to file
                    method="http" -> ingest using HTTP POST requests
                    method="kafka" -> ingest to kafka endpoint
        :param target: string
            Used to identify where the data should be ingested into.
                The value for this parameter depends on the ingest method chosen.
                For "http" method, it would require an HTTP URL.
                    Example: http://example.com/<some>/<api>/<endpoint>
                For "file" method, it would require a file name or path.

                See chosen ingest method (ingestion plugin) documentation for more details on the expected target format.
        :param collection_id: string
            (Optional) An identifier to indicate that data from different method
            invocations belong to same collection. Defaults to "data_job_name|OpID",
            meaning all method invocations from a data job run will belong to the same collection.
        """
        self._objects_queue.put(
            (payload_dict, destination_table, method, target, collection_id)
        )

    def _payload_aggregator_thread(self):
        """
        Thread aggregating the ingestion data based on the provided payload size
        threshold and target/collection_id
        """
        aggregated_payload = []
        number_of_payloads = 0
        current_payload_size_in_bytes = 0
        current_target = None
        current_collection_id = None
        current_destination_table = None
        method = None
        while self._closed.value == 0:
            try:
                (
                    payload_dict,
                    destination_table,
                    method,
                    target,
                    collection_id,
                ) = self._objects_queue.get(
                    timeout=self._payload_aggregator_timeout_seconds
                )
            except queue.Empty:
                (
                    aggregated_payload,
                    number_of_payloads,
                    current_payload_size_in_bytes,
                ) = self._queue_payload_for_posting(
                    aggregated_payload,
                    number_of_payloads,
                    destination_table,
                    method,
                    current_target,
                    current_collection_id,
                )
                continue

            # First payload will determine the target and collection_id
            if (
                not current_target
                and not current_collection_id
                and not current_destination_table
            ):
                current_target = target
                current_collection_id = collection_id
                current_destination_table = destination_table

            # When we get a payload with different than current target/collection_id/destination_table,
            # send the current payload and start aggregating for the new one.
            if (
                current_target != target
                or current_collection_id != collection_id
                or current_destination_table != destination_table
            ):
                (
                    aggregated_payload,
                    number_of_payloads,
                    current_payload_size_in_bytes,
                ) = self._queue_payload_for_posting(
                    aggregated_payload,
                    number_of_payloads,
                    current_destination_table,
                    method,
                    current_target,
                    current_collection_id,
                )
                current_target = target
                current_collection_id = collection_id
                current_destination_table = destination_table

            # We are converting to string to get correct memory size. This may
            # cause performance issues.
            # TODO: Propose a way to calculate the object's memory footprint without converting to string.
            string_payload = str(payload_dict)

            if (
                sys.getsizeof(string_payload) + current_payload_size_in_bytes
                > self._payload_size_bytes_threshold
            ):
                (
                    aggregated_payload,
                    number_of_payloads,
                    current_payload_size_in_bytes,
                ) = self._queue_payload_for_posting(
                    aggregated_payload,
                    number_of_payloads,
                    current_destination_table,
                    method,
                    current_target,
                    current_collection_id,
                )
            aggregated_payload.append(payload_dict)
            number_of_payloads += 1
            current_payload_size_in_bytes += sys.getsizeof(string_payload)

    def _queue_payload_for_posting(
        self,
        aggregated_payload: list,
        number_of_payloads: int,
        destination_table: str,
        method: str,
        target: str,
        collection_id: str,
    ):
        """
        Send payload to the _payloads_queue.

        :param aggregated_payload: list
            List of aggregated payloads that are ready for final processing.
        :param number_of_payloads: int,
            Number of payloads to be dequeued from the _objects_queue.
        :param destination_table: string
            The name of the table, where the data should be ingested into.
        :param method: string
            Indicates the ingestion method to be used. E.g.:
                    method="file" -> ingest to file
                    method="http" -> ingest using HTTP POST requests
                    method="kafka" -> ingest to kafka endpoint
        :param collection_id: string
            An identifier to indicate that data from different method
            invocations belong to same collection. Defaults to "data_job_name|OpID",
            meaning all method invocations from a data job run will belong to the
            same collection.
        :param target: string
            Used to identify where the data should be ingested into.
                The value for this parameter depends on the ingest method chosen.
                For "http" method, it would require an HTTP URL.
                    Example: http://example.com/<some>/<api>/<endpoint>
                For "file" method, it would require a file name or path.

                See chosen ingest method (ingestion plugin) documentation for more details on the expected target format.
        """
        if aggregated_payload:
            try:
                self._payloads_queue.put(
                    (
                        aggregated_payload,
                        destination_table,
                        method,
                        target,
                        collection_id,
                    )
                )
            except Exception as e:
                self._fail_count.increment()
                if self._log_upload_errors:
                    log.warning(
                        "Failed to send a payload for ingestion. One or more rows were not ingested.\n"
                        "Exception was: {}".format(str(e))
                    )
            finally:
                for i in range(number_of_payloads):
                    self._objects_queue.task_done()
        # Return statement below is coupled with _payload_aggregator_thread
        # and is used to reset the aggregated payload, number of payloads and
        # its aggregated payload size
        aggregated_payload_reset: list = []
        number_of_payloads_reset: int = 0
        current_payload_size_in_bytes_reset: int = 0
        return (
            aggregated_payload_reset,
            number_of_payloads_reset,
            current_payload_size_in_bytes_reset,
        )

    def _payload_poster_thread(self):
        """
        Thread doing final data preparation (adding additional metadata, telemetry
        data, etc.) and ingesting the data.
        """
        while self._closed.value == 0:
            try:
                ingestion_metadata: Optional[IIngesterPlugin.IngestionMetadata] = None
                exception: Optional[Exception] = None
                payload_obj: Optional[List] = None
                destination_table: Optional[str] = None
                target: Optional[str] = None
                collection_id: Optional[str] = None
                try:
                    payload = self._payloads_queue.get()
                    payload_obj, destination_table, _, target, collection_id = payload

                    # If there are any pre-processors set, pass the payload object
                    # through them.
                    if self._pre_processors:
                        payload_obj, ingestion_metadata = self._pre_process_payload(
                            payload=payload_obj,
                            destination_table=destination_table,
                            target=target,
                            collection_id=collection_id,
                            metadata=ingestion_metadata,
                        )

                    try:
                        ingestion_metadata = self._ingester.ingest_payload(
                            payload=payload_obj,
                            destination_table=destination_table,
                            target=target,
                            collection_id=collection_id,
                            metadata=ingestion_metadata,
                        )

                        self._success_count.increment()
                    except VdkConfigurationError:
                        # TODO: logging for every error might be too much
                        # There could be million of uploads and millions of error logs would be hard to use.
                        # But until we have a way to aggregate the errors and show
                        # the most relevant errors it's better to make sure we do not hide an issue
                        # and be verbose.
                        log.exception(
                            "A configuration error occurred while ingesting data."
                        )
                        raise
                    except UserCodeError:
                        log.exception("An user error occurred while ingesting data.")
                        raise
                    except Exception:
                        log.exception("A platform error occurred while ingesting data.")
                        raise

                except Exception as e:
                    self._fail_count.increment()
                    if self._log_upload_errors:
                        # TODO: When working on row count telemetry we can add the exact number of rows not ingested.
                        log.warning(
                            "Failed to ingest a payload. "
                            "One or more rows were not ingested.\n"
                            "Exception was: {}".format(str(e))
                        )
                    exception = e
                finally:
                    self._payloads_queue.task_done()

                # If there are any post-processors set, complete the post-process
                # operations
                if self._post_processors:
                    self._execute_post_process_operations(
                        payload=payload_obj,
                        destination_table=destination_table,
                        target=target,
                        collection_id=collection_id,
                        metadata=ingestion_metadata,
                        exception=exception,
                    )
            except VdkConfigurationError:
                self._plugin_errors[VdkConfigurationError].increment()
            except UserCodeError:
                self._plugin_errors[UserCodeError].increment()
            except Exception:
                self._plugin_errors[PlatformServiceError].increment()

    def _start_workers(self):
        """
        Start the worker threads.
        """
        t = threading.Thread(
            target=self._payload_aggregator_thread, name="payload-aggregator"
        )
        t.daemon = True
        t.start()
        for i in range(self._number_of_worker_threads):
            thread_name = f"payload-poster{i}"
            t = threading.Thread(target=self._payload_poster_thread, name=thread_name)
            t.daemon = True
            t.start()

    def __wait_if_necessary(self):
        if self._wait_to_finish_after_every_send:
            self.__wait_to_finish()

    def __wait_to_finish(self):
        """
        Wait for completion of processing of all data added to the payload
        queue.
        """
        ingester_utils.wait_completion(
            objects_queue=self._objects_queue, payloads_queue=self._payloads_queue
        )

    def close(self):
        """
        Wait for completion of processing of all data added to the payload
        queue and then log the ingestion statistics.
        """
        self.__wait_to_finish()
        self.close_now()

    def close_now(self):
        """
        Close immediately. The method will not wait for the active queue items to get processed.
        """
        if self._closed.get_and_increment() == 0:

            if self._exception_on_failure:
                self.__handle_results()

            log.info(
                "Ingester statistics: \n\t\t"
                f"Successful uploads:{self._success_count}\n\t\t"
                f"Failed uploads:{self._fail_count}\n\t\t"
                f"ingesting plugin errors:{self._plugin_errors}\n\t\t"
            )

    def _pre_process_payload(
        self,
        payload: List[dict],
        destination_table: Optional[str],
        target: Optional[str],
        collection_id: Optional[str],
        metadata: Optional[IIngesterPlugin.IngestionMetadata],
    ) -> Tuple[List[dict], Optional[IIngesterPlugin.IngestionMetadata]]:
        for plugin in self._pre_processors:
            try:
                payload, metadata = plugin.pre_ingest_process(
                    payload=payload,
                    destination_table=destination_table,
                    target=target,
                    collection_id=collection_id,
                    metadata=metadata,
                )
            except Exception as e:
                errors.log_and_throw(
                    to_be_fixed_by=ResolvableBy.USER_ERROR,
                    log=log,
                    what_happened="Failed to pre-process the data.",
                    why_it_happened=f"User Error occurred. Exception was: {e}",
                    consequences="Execution of the data job will fail, "
                    "in order to prevent data corruption.",
                    countermeasures="Check if the data sent for ingestion "
                    "is aligned with the requirements, "
                    "and that the pre-process plugins are "
                    "configured correctly.",
                )
        return payload, metadata

    def _execute_post_process_operations(
        self,
        payload: List[dict],
        destination_table: Optional[str],
        target: Optional[str],
        collection_id: Optional[str],
        metadata: Optional[IIngesterPlugin.IngestionMetadata],
        exception: Optional[Exception],
    ):
        for plugin in self._post_processors:
            try:
                metadata = plugin.post_ingest_process(
                    payload=payload,
                    destination_table=destination_table,
                    target=target,
                    collection_id=collection_id,
                    metadata=metadata,
                    exception=exception,
                )
            except Exception as e:
                errors.log_and_throw(
                    to_be_fixed_by=ResolvableBy.USER_ERROR,
                    log=log,
                    what_happened="Could not complete post-ingestion process.",
                    why_it_happened=f"User Error occurred. Exception was: {e}",
                    consequences="Execution of the data job will fail, "
                    "in order to prevent data corruption.",
                    countermeasures="Check if the data sent for "
                    "post-processing "
                    "is aligned with the requirements, "
                    "and that the post-process plugins are "
                    "configured correctly.",
                )

    def __handle_results(self):
        if self._plugin_errors.get(UserCodeError, AtomicCounter(0)).value > 0:
            self._log_and_throw(
                to_be_fixed_by=ResolvableBy.USER_ERROR,
                countermeasures="Ensure data you are sending is aligned with the requirements",
                why_it_happened="User error occurred. See warning logs for more details. ",
            )
        if self._plugin_errors.get(VdkConfigurationError, AtomicCounter(0)).value > 0:
            self._log_and_throw(
                to_be_fixed_by=ResolvableBy.CONFIG_ERROR,
                countermeasures="Ensure job is properly configured. "
                "For example make sure that target and method specified are correct",
            )
        if (
            self._plugin_errors.get(PlatformServiceError, AtomicCounter(0)).value > 0
            or self._fail_count.value > 0
        ):
            self._log_and_throw(
                to_be_fixed_by=ResolvableBy.PLATFORM_ERROR,
                countermeasures="There has been temporary failure. "
                "You can retry the data job again. "
                "If error persist inspect logs and check ingest plugin documentation.",
            )

    def _log_and_throw(
        self,
        to_be_fixed_by,
        countermeasures,
        why_it_happened="Payloads failed to get posted for ingestion",
    ):
        errors.log_and_throw(
            to_be_fixed_by=to_be_fixed_by,
            log=log,
            what_happened="Failed to post all data for ingestion successfully.",
            why_it_happened=why_it_happened,
            consequences="Some data will not be ingested.",
            countermeasures=countermeasures,
        )

    @staticmethod
    def __verify_payload_format(payload_dict: dict):
        if not payload_dict:
            raise errors.UserCodeError(
                "Payload given to " "ingestion method should " "not be empty."
            )

        elif not isinstance(payload_dict, dict):
            raise errors.UserCodeError(
                "Payload given to ingestion method should be a "
                "dictionary, but it is not."
            )

        # Check if payload dict is valid json
        # TODO: optimize the check - we should not need to serialize the payload every time
        try:
            json.dumps(payload_dict)
        except (TypeError, OverflowError, Exception) as e:
            errors.log_and_throw(
                errors.ResolvableBy.USER_ERROR,
                log,
                "Failed to send payload",
                "JSON Serialization Error. Payload is not json serializable",
                "Payload may be only partially ingested, or not ingested at all.",
                f"See error message for help: {str(e)}",
            )
